#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <stdint.h>
#include <unistd.h>

typedef uint64_t u64;
typedef int64_t s64;
typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;
typedef struct {
	int counter;
} atomic_t;
#define ATOMIC_INIT(i) { (i) }
typedef struct {
	s64 counter;
} atomic64_t;
#include "linux/arch/arm64/include/asm/atomic_ll_sc.h"

atomic_t atomic_var;
char separator[1000000];
atomic_t atomic_var_uncontested;
#ifdef UNCONTESTED
    #define LOAD_THREAD_ATOMIC_VAR atomic_var_uncontested
#else
    #define LOAD_THREAD_ATOMIC_VAR atomic_var
#endif

#define NANOSEC_PER_SEC 1000000000 /* 10^9 */
static uint64_t gettime(void)
{
    struct timespec t;
    int res;

    res = clock_gettime(CLOCK_TAI, &t);
    if (res < 0) {
        fprintf(stderr, "Error with gettimeofday! (%i)\n", res);
        exit(1);
    }
    return (uint64_t) t.tv_sec * NANOSEC_PER_SEC + t.tv_nsec;
}

void *load_thread_function( void *ptr )
{
     printf("load thread started\n");
     for(;;) {
         __ll_sc_atomic_fetch_xor_acquire(1, &LOAD_THREAD_ATOMIC_VAR);
     }
     
}

void *evaluation_thread_function( void *ptr )
{
     printf("evaluation thread started \n");
     uint64_t max_duration = 0;
     for(;;) {
         uint64_t start = gettime();
         __ll_sc_atomic_fetch_or_acquire(4, &atomic_var);
         uint64_t end = gettime();
         uint64_t duration = end - start;
         if (duration > max_duration) {
             max_duration = duration;
             printf("new max duration: %lu ns\n", max_duration);
         }
     }
}

int main()
{
     pthread_t load_thread;
     pthread_t evaluation_thread;

     pthread_create( &load_thread, NULL, load_thread_function, NULL);
     sleep(1);
     pthread_create( &evaluation_thread, NULL, evaluation_thread_function, NULL);

     pthread_join( load_thread, NULL);
     pthread_join( evaluation_thread, NULL);
}

